MAJOR_VERSION = 1
MINOR_VERSION = 2
PATCH_VERSION = 28
version = '{}.{}.{}'.format(MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION)
__all__  = ['MAJOR_VERSION', 'MINOR_VERSION', 'PATCH_VERSION', 'version']



#!/usr/bin/python3

## script to load results from a basic dataloading script

import os
import sys
from mb_pytorch.dataloader.loader import DataLoader
from mb_utils.src.logging import logger
import argparse



def main():
    data = args.file

    k =DataLoader('./scripts/loader_y.yaml',logger=logger) #load the data

    if logger:
        logger.info("Data loaded")
        logger.info("Data params: {}".format(k.load_data_params))

    out1,out2,o1,o2 =k.data_load(logger=logger)
    if logger:
        #logger.info("self.trainset [0] = {}".format(o1[0]))
        logger.info("self.trainset keys = {}".format(o1[0].keys()))
        logger.info("self.trainset shape = {}".format(o1[0]['image'].shape))

    for i in out1:
        if logger:
            #logger.info("self.trainloader [0] = {}".format(i['image']))
            logger.info("self.trainloader keys = {}".format(i.keys()))
            logger.info("self.trainloader shape = {}".format(i['image'].shape))
            if 'label' in i.keys():
                logger.info("self.trainloader label = {}".format(i['label']))
        break


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--file',type=str,help='path to yaml data file',default='./scripts/loader_y.yaml')
    args = parser.parse_args()
    main()



##Class for generating embeddings for the given data

import torch
from mb_pytorch.dataloader.loader import DataLoader
import cv2
import numpy as np
from tqdm import tqdm
from mb_pandas.src.dfload import load_any_df
from mb_utils.src.verify_image import verify_image
from mb_pandas.src.transform import *
import os
import torchvision
import torch.nn as nn
from torch.utils import hooks


__all__ = ['EmbeddingGenerator']


class customdl_emb(torch.utils.data.Dataset):
    def __init__(self,data,transform=None,train_file=True,logger=None):
        self.transform=transform
        self.logger=logger
        self.folder_name=data['work_dir']
        self.data = load_any_df(data['file'],logger=self.logger)

        if self.logger:
            self.logger.info("Data file: {} loaded with mb_pandas.".format(data))
            self.logger.info("Data columns: {}".format(self.data.columns))
            self.logger.info("Data will be split into train and validation according to train_file input : {}".format(train_file))
            self.logger.info("If unnamed columns are present, they will be removed.")
            self.logger.info("If duplicate rows are present, they will be removed.")
        assert 'image_path' in self.data.columns, "image_path column not found in data"

        self.data = check_drop_duplicates(self.data,columns=['image_path'],drop=True,logger=self.logger)
        self.data = remove_unnamed(self.data,logger=self.logger)

        if data['use_img_dir']:
            img_path = [os.path.join(str(data['img_dir']),self.data['image_path'].iloc[i]) for i in range(len(self.data))]
        else:
            img_path = [self.data['image_path'].iloc[i] for i in range(len(self.data))]
        self.data['image_path_new'] = img_path
        if self.logger:
            self.logger.info("Verifying paths")
            self.logger.info("first path : {}".format(img_path[0]))

        path_check_res= [os.path.exists(img_path[i]) for i in range(len(img_path))]
        self.data['img_path_check'] = path_check_res
        self.data = self.data[self.data['img_path_check'] == True]
        self.data = self.data.reset_index(drop=True)
        if logger:
            self.logger.info("self.data: {}".format(self.data))
            self.logger.info("Length of data after removing invalid paths: {}".format(len(self.data)))
            self.logger.info("Verifying images")

        verify_image_res = [verify_image(self.data['image_path_new'].iloc[i],logger=self.logger) for i in range(len(self.data))]  
        self.data['img_verify'] = verify_image_res
        self.data = self.data[self.data['img_verify'] == True]
        self.data = self.data.reset_index()
        
        if os.path.exists(self.folder_name):
            self.data.to_csv(os.path.join(self.folder_name,'emb_wrangled_file.csv'),index=False)

    def __len__(self):
        return len(self.data)
    
    def __repr__(self) -> str:
        return "self.data: {},self.transform: {}".format(self.data,self.transform)

    def __getitem__(self,idx):
        
        img = self.data['image_path_new'].iloc[idx]
        #img = Image.open(img)
        img = cv2.imread(img)

        if self.transform:
            img = self.transform(img)

        out_dict = {'image':img}                                           
        return out_dict
    
class EmbeddingGenerator(DataLoader):
    def __init__(self, yaml, logger=None) -> None:
        super().__init__(yaml,logger=logger)
        
        self._data = self.load_data_all
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.data_file = self._data['data']['file']
        self.model = self._data['model']['model']
        self.use_pretrained = self._data['model']['use_pretrained']
        self.use_own_model = self._data['model']['use_own_model']
        if self.use_own_model:
            self.model = torch.load(self._data['model']['model_path'])

        self.ext_layer = self._data['model']['model_layer']
        self.transforms_final = self.get_transforms
        self._emb = None
        self.logger = logger

    def model_set(self,name=None):
        """
        Set the model for embedding generation. model already set in the yaml file
        """
        if name:
            k=eval("torchvision.models."+name)
        else:
            k=eval("torchvision.models."+self.model)
        self.model_val = k(pretrained=True)
        if self.logger:
            self.logger.info("Model set to {}".format(self._data['model']['model']))
        return self.model_val
    
    def generate_emb(self, data):
        """
        Generate embeddings for the given data
        Input:
            data: data for which embeddings are to be generated (numpy array)
        Output:
            emb: embeddings for the given data
        """
        model = self.model_set()
        model.to(self.device)
        model.eval()
        
        features_blobs = []
        def get_hook(module, input, output):
            N,C,H,W = output.shape
            output = output.reshape(N,C,-1)
            features_blobs.append(output.data.cpu().numpy())

        fea = model._modules.get(self.ext_layer).register_forward_hook(get_hook)
    
        if self.logger:
            self.logger.info("Embedding generation started")
            self.logger.info("length of data: {}".format(len(data)))

        for i,i_dat in tqdm(enumerate(data),total=len(data)):
            _ = model(i_dat['image'].to(self.device))

        self.emb= np.concatenate(features_blobs)
        self.emb = self.emb.reshape(self.emb.shape[0],self.emb.shape[1])

        fea.remove()
        del model
        if self.logger:
            self.logger.info("Embedding generation completed")
        return self.emb
    
    def file_save(self,logger=None):
        """
        Save the embeddings to a wrnagled file
        """
        work_dir = self._data['data']['work_dir']
        if os.path.exists(self.folder_name):
            df = load_any_df(os.path.join(self.folder_name,'emb_wrangled_file.csv'))
        df['embedding'] = self.emb.tolist()
        df.to_csv(os.path.join(self.folder_name,'emb_wrangled_file.csv'),index=False)
        if self.logger:
            self.logger.info("Embeddings saved to {}".format(os.path.join(self.folder_name,'emb_wrangled_file.csv')))


    def data_emb_loader(self):
        """
        get embedding data from yaml file
        """
        data = self.data_train_emb(self._data['data'],transform=self.get_transforms,train_file=False,logger=self.logger)
        loader = torch.utils.data.DataLoader(data, batch_size=self._data['train_params']['batch_size'],shuffle=self._data['train_params']['shuffle'],
                                              num_workers=self._data['train_params']['num_workers'])
        return loader

    def data_train_emb (self,data_file,transform,train_file,logger=None):
        """
        get train data from yaml file
        """
        data_t = customdl_emb(data_file,transform=transform,train_file=train_file,logger=logger)
        return data_t

        



import torch
from typing import Optional
import tqdm
import os
from mb_utils.src.logging import logger
import numpy as np
from ..utils.viewer import gradcam_viewer,plot_classes_pred,plot_to_image
from mb_pytorch.models.modelloader import ModelLoader
from mb_pytorch.training.train_params import train_helper
from mb.plt.utils import dynamic_plt

__all__ = ['detection_train_loop']

def detection_train_loop( k_yaml: dict,scheduler: Optional[object] =None,writer: Optional[object] =None,
                              logger: Optional[object] =None,gradcam: Optional[object] =None,
                              gradcam_rgb: str =False,device: str ='cpu'):
    """
    Function to train the model
    Args:
        k_yaml: data dictionary YAML of DataLoader
        scheduler: scheduler
        writer: tensorboard writer
        logger: logger
        gradcam: gradcam layers to be visulized
        device: default is cpu
    output:
        None
    """
    
    if logger:
        logger.info('Training loop Starting')
    k_data = k_yaml.data_dict['data']
    data_model = k_yaml.data_dict['model']
    model_data_load = ModelLoader(k_yaml.data_dict['model'])
    model =  model_data_load.get_model()
    device_type = data_model['device']
    bbox_threshold = data_model['model_meta_data']['model_bbox_threshold']

    if logger:
        logger.info('Model Loaded')
    
    train_loader,val_loader,_,_ = k_yaml.data_load()
    loss_attr,optimizer_attr,optimizer_dict,scheduler_attr,scheduler_dict = train_helper(data_model) 
    optimizer = optimizer_attr(model.parameters(),**optimizer_dict)
    if scheduler is not None:
        scheduler = scheduler_attr(optimizer,**scheduler_dict)

    if logger:
        logger.info('Optimizer and Scheduler Loaded')
        logger.info(f'Loss: {loss_attr}')
        logger.info(f'Optimizer: {optimizer}')
        logger.info(f'Scheduler: {scheduler}')

    if device_type =='cuda':
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        if device == 'cuda':
            torch.cuda.empty_cache()
            if logger:
                logger.info('Device: CUDA')
                logger.info(torch.cuda.get_device_name(0))

    model.to(device)
    best_val_loss = float('inf')

    for epoch in tqdm.tqdm(range(data_model['model_epochs']), desc="Epochs"):
        
        ##train loop
        
        model.train(True)
        train_loss = 0
        
        if logger:
            logger.info('Training Started')
        for batch_idx, data in enumerate(tqdm.tqdm(train_loader, desc="Training", leave=False)):
            images,bbox,labels = data.values()
            images = list(image.to(device) for image in images)
            bbox = list(b.to(device) for b in bbox)
            bbox = [b.view(-1, 4) if b.dim() == 1 else b for b in bbox]
            labels = list(label.to(device) for label in labels)  
            targets = [{'boxes': b,'labels': label} for b,label in zip(bbox, labels)]      
                    
            optimizer.zero_grad()
            loss_dict = model(images, targets)  
            losses = sum(loss for loss in loss_dict.values())
            
            losses.backward()
            optimizer.step()
            if scheduler is not None:
                scheduler.step()
            
            train_loss += losses.item()
            if logger:
                logger.info(f'Epoch {epoch+1} - Batch {batch_idx+1} - Train Loss: {losses.item()}')
        
        avg_train_loss = train_loss / len(train_loader)
        if logger:
            logger.info(f'Epoch {epoch+1} - Train Loss: {avg_train_loss}')
            logger.info(f"lr = {optimizer.param_groups[0]['lr']}")

        model.train(False)

        ## Validation loop
        val_loss = 0
        val_bbox = []
        val_labels = []
        val_scores = []
        val_targets_labels = []
        val_targets_bbox = []

        with torch.no_grad():
            for batch_idx, data in enumerate(tqdm.tqdm(val_loader, desc="Validation", leave=False)):
                images,bbox,labels = data.values()
                images = list(image.to(device) for image in images)
                bbox = list(b.to(device) for b in bbox)
                bbox = [b.view(-1, 4) if b.dim() == 1 else b for b in bbox]
                labels = list(label.to(device) for label in labels)  
                targets = [{'boxes': b,'labels': label} for b,label in zip(bbox, labels)]    

                loss_dict = model(images, targets)  ## output is boxes, labels, scores

                if len(loss_dict) == 0:
                    continue
                else:
                    for i in range(len(loss_dict)):
                        if len(loss_dict[i]['boxes']) > 0 :
                            if 'scores' in loss_dict[i]:
                                for j in range(len(loss_dict[i]['scores'])):
                                    if loss_dict[i]['scores'][j] > bbox_threshold:
                                        val_bbox.append(loss_dict[i]['boxes'][j])
                                        val_labels.append(loss_dict[i]['labels'][j])
                                        val_scores.append(loss_dict[i]['scores'][j])
                                        if j < 1:
                                            val_targets_labels.append(labels[i])
                                            val_targets_bbox.append(bbox[i])

                #losses = sum(loss for loss in loss_dict.values() if loss.numel()>0)
                
                val_loss += losses.item() * len(images)
                if logger: 
                    logger.info(f'Epoch {epoch+1} - Batch {batch_idx+1} - Val Loss: {losses.item()}')
            
            avg_val_loss = val_loss / len(val_loader.dataset)
            
            if logger:
                logger.info(f'Epoch {epoch+1} - Avg Val Loss: {avg_val_loss:.3f}')

        # TensorBoard logging
        if writer is not None:
            writer.add_scalar('Loss/train', avg_train_loss, global_step=epoch)
            writer.add_scalar('Loss/val', avg_val_loss, global_step=epoch)
            writer.add_scalar('Learning_rate', optimizer.param_groups[0]['lr'], global_step=epoch)
            
            for name, param in model.named_parameters():
                writer.add_histogram(name, param, global_step=epoch)
            
            # Visualizations
            if len(images) > 0:
                img_l = [np.array(i.to('cpu')) for i in images]
                labels_l = [str(list(np.array((l.to('cpu'))))[0]) for l in labels] ##check for multiple labels
                #create_img_grid(x, x, writer, global_step=epoch)
                fig = dynamic_plt(img_l,labels=labels_l,return_fig=True)
                writer.add_image('grid', plot_to_image(fig), global_step=epoch)

                # Grad-CAM visualization 
                if gradcam is not None:
                    use_cuda = device != 'cpu'
                    for cam_layers in gradcam:
                        grad_img = gradcam_viewer(cam_layers, model, x.unsqueeze(0), gradcam_rgb=gradcam_rgb, use_cuda=use_cuda) 
                        if grad_img is not None:
                            grad_img = np.transpose(grad_img, (2, 0, 1))
                            writer.add_image(f'Gradcam/{cam_layers}', grad_img, global_step=epoch)
                        elif logger:
                            logger.info(f'Gradcam not supported for {cam_layers}')   
   
        # Save best model
        if avg_val_loss < best_val_loss:
            best_val_loss = avg_val_loss
            best_model = model.state_dict()

            path = os.path.join(os.path.dirname(k_data['file']['root']), 'best_model.pth')
            torch.save(best_model, path)
            if logger:
                logger.info(f'Epoch {epoch + 1} - Best Model Saved (Val Loss: {best_val_loss:.4f})')


import torch.nn.functional as F

__all__ = ['dice_loss', 'binary_cross_entropy', 'cross_entropy', 'mse_loss', 'l1_loss']

def dice_loss(input, target):
    """
    Dice loss function
    Formula: 1 - (2 * intersection + smooth) / (union + smooth)
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        dice_loss: torch tensor of shape (1)
    """
    smooth = 1.
    iflat = input.view(-1)
    tflat = target.view(-1)
    intersection = (iflat * tflat).sum()
    return 1 - ((2. * intersection + smooth) /
                (iflat.sum() + tflat.sum() + smooth))

def binary_cross_entropy(input, target):
    """
    Binary cross entropy loss function
    Formula: -sum(y * log(x) + (1 - y) * log(1 - x))
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        binary_cross_entropy_loss: torch tensor of shape (1)
    """
    return F.binary_cross_entropy(input, target)

def cross_entropy(input, target):
    """
    Cross entropy loss function
    Formula: -sum(y * log(x))
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        cross_entropy_loss: torch tensor of shape (1)
    """
    return F.cross_entropy(input, target)

def mse_loss(input, target):
    """
    Mean squared error loss function
    Formula: 1/N * sum((x - y)^2)
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        mse_loss: torch tensor of shape (1)
    """
    return F.mse_loss(input, target)

def l1_loss(input, target):
    """
    L1 loss function
    Formula: |x - y|
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        l1_loss: torch tensor of shape (1)
    """
    return F.l1_loss(input, target)

def hinge_loss(input, target):
    """
    Hinge loss function
    Formula: max(0, 1 - y * x)
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        hinge_loss: torch tensor of shape (1)
    """
    return F.hinge_embedding_loss(input, target)

def l2_loss(input, target):
    """
    L2 loss function
    Formula: (x - y)^2
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        l2_loss: torch tensor of shape (1)
    """
    return F.mse_loss(input, target)

def compute_iou(box1, box2):
    """
    Compute the Intersection over Union (IoU) of two bounding boxes.
    
    Args:
        box1 (torch.Tensor): The first bounding box, of the form [x1, y1, x2, y2].
        box2 (torch.Tensor): The second bounding box, of the form [x1, y1, x2, y2].
    
    Returns:
        iou (float): The Intersection over Union of the two boxes.
    """
    x1 = max(box1[0], box2[0])
    y1 = max(box1[1], box2[1])
    x2 = min(box1[2], box2[2])
    y2 = min(box1[3], box2[3])
    
    intersection = max(0, x2 - x1 + 1) * max(0, y2 - y1 + 1)
    
    box1_area = (box1[2] - box1[0] + 1) * (box1[3] - box1[1] + 1)
    box2_area = (box2[2] - box2[0] + 1) * (box2[3] - box2[1] + 1)
    
    iou = intersection / float(box1_area + box2_area - intersection)
    
    return iou


def compute_loss(outputs, targets):
    """
    Compute the loss for a FasterRCNN model with a ResNet backbone.
    
    Args:
        outputs (dict): A dictionary containing the model's outputs, including 'boxes', 'labels', and 'scores'.
        targets (dict): A dictionary containing the ground truth bounding boxes and labels.
    
    Returns:
        loss (torch.Tensor): The total loss.
    """
    # Extract the relevant outputs and targets
    predicted_boxes = outputs['boxes']
    predicted_labels = outputs['labels']
    
    gt_boxes = targets['boxes']
    gt_labels = targets['labels']
    
    # Compute the classification loss
    classification_loss = F.cross_entropy(predicted_labels, gt_labels)
    
    # Compute the bounding box regression loss
    num_predicted = predicted_boxes.size(0)
    num_ground_truth = gt_boxes.size(0)
    
    bbox_regression_loss = 0
    for i in range(num_predicted):
        # Find the best matching ground truth box
        best_iou = 0
        best_index = -1
        for j in range(num_ground_truth):
            iou = compute_iou(predicted_boxes[i], gt_boxes[j])
            if iou > best_iou:
                best_iou = iou
                best_index = j
        
        # Compute the smooth L1 loss for the best matching box
        if best_index >= 0:
            bbox_regression_loss += F.smooth_l1_loss(predicted_boxes[i], gt_boxes[best_index])
    
    bbox_regression_loss /= num_predicted
    
    # Compute the total loss
    total_loss = classification_loss + bbox_regression_loss
    
    return total_loss ##look into adding boundingbox + classification loss for validation


#dataloader for pytorch1.0

import torch
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
from mb_pytorch.utils.yaml_reader import YamlReader
import os
import numpy as np
import pandas as pd
from mb_pandas.src.dfload import load_any_df
from mb_utils.src.verify_image import verify_image
from mb.pandas import check_drop_duplicates,remove_unnamed
from ..utils.extra_utils import labels_num_map
from datetime import datetime
import cv2

today = datetime.now()

__all__ = ['data_fetcher','DataLoader']

class data_fetcher:
    """
    dataloader for pytorch1.0
    """
    def __init__(self,yaml,logger=None) -> None:
        self.yaml = yaml
        self.logger = logger
        self._yaml_data = None
        self.data_dict = {}
        self.transforms_final=[]
        self.all = None

    def __repr__(self) -> str:
        return "data_fetcher(yaml={},logger={})".format(self.yaml,self.logger)

    @staticmethod
    def read_yaml(self):
        """
        read yaml file
        """
        self._yaml_data = YamlReader(self.yaml).data(self.logger)
        return self._yaml_data

    @property
    def load_data_params(self):
        """
        get dataloader data from yaml file
        """
        data = YamlReader(self.yaml).data(self.logger)
        self.data_dict['data'] = data['data']
        self.data_dict['train_params'] = data['train_params']
        self.data_dict['test_params'] = data['test_params']
        self.data_dict['transformation'] = data['transformation']
        self.data_dict['model'] = data['model']
        return self.data_dict
    
    @property
    def load_data_all(self):
        """
        get dataloader all data dict from yaml file
        """
        data = YamlReader(self.yaml).data(self.logger)
        self.all = data
        return self.all


class JointTransforms:
    def __init__(self,transform_yaml,logger=None):
        """
        get transforms from yaml file
        """

        self.transform_data = transform_yaml
        self.logger = logger

        if self.transform_data['transform']==False:  # noqa: E712
            return None

    def __call__(self,img,mask=None,bbox=None):
        if self.transform_data['to_tensor']['val']:
            img = transforms.ToTensor()(img)
            if mask is not None:
                mask = transforms.ToTensor()(mask)
            if bbox is not None:
                bbox = torch.tensor([bbox[0],bbox[1],bbox[2],bbox[3]],dtype=torch.int32)

        if self.transform_data['normalize']['val']:
            img = transforms.Normalize(self.transform_data['normalize']['args']['mean'],self.transform_data['normalize']['args']['std'])(img)

        if self.transform_data['resize']['val']:
            ori_size = img.size()
            img = transforms.Resize(self.transform_data['resize']['args']['size'])(img)
            if mask is not None:
                mask = transforms.Resize(self.transform_data['resize']['args']['size'])(mask)
            if bbox is not None:
                bbox = self.resize_boxes(ori_size,img.size(),bbox)

        if self.transform_data['random_crop']['val']:
            img = transforms.RandomCrop(self.transform_data['random_crop']['args']['size'])(img)
            if mask is not None:
                mask = transforms.RandomCrop(self.transform_data['random_crop']['args']['size'])(mask)
            if bbox is not None:
                bbox = self.crop_boxes(bbox, *self.transform_data['random_crop']['args']['size'])

        if self.transform_data['random_horizontal_flip']['val']:
            img = transforms.RandomHorizontalFlip(self.transform_data['random_horizontal_flip']['args']['p'])(img)
            if mask is not None:
                mask = transforms.RandomHorizontalFlip(self.transform_data['random_horizontal_flip']['args']['p'])(mask)
            if bbox is not None:
                bbox = self.hflip_boxes(bbox, img.size()[1])

        if self.transform_data['random_vertical_flip']['val']:
            img = transforms.RandomVerticalFlip(self.transform_data['random_vertical_flip']['args']['p'])(img)
            if mask is not None:
                mask = transforms.RandomVerticalFlip(self.transform_data['random_vertical_flip']['args']['p'])(mask)
            if bbox is not None:
                bbox = self.vflip_boxes(bbox, img.size()[2])

        if self.transform_data['random_rotation']['val']:
            img = transforms.RandomRotation(self.transform_data['random_rotation']['args']['degrees'])(img)
            if mask is not None:
                mask = transforms.RandomRotation(self.transform_data['random_rotation']['args']['degrees'])(mask)
            if bbox is not None:
                bbox = self.rotate_boxes(bbox, self.transform_data['random_rotation']['args']['degrees'], img.size[1], img.size[0])

        if self.transform_data['random_color_jitter']['val']:
            img = transforms.ColorJitter(brightness=self.transform_data['random_color_jitter']['args']['brightness'],contrast=self.transform_data['random_color_jitter']['args']['contrast'],saturation=self.transform_data['random_color_jitter']['args']['saturation'],hue=self.transform_data['random_color_jitter']['args']['hue'])(img)
        
        if self.transform_data['random_grayscale']['val']:
            img = transforms.RandomGrayscale(self.transform_data['random_grayscale']['args']['p'])(img)
        
        if mask is not None:
            return img,mask
        elif bbox is not None:
            return img,bbox
        else:
            return img
        
    def resize_boxes(self,original_size, new_size, bbox):
        """
        Resize bounding boxes according to the new image size
        Input:
        original_size: torch.tensor, (channel, original_height, original_width)
        new_size: tuple, (channel, new_height,new_width)
        bbox: list, [x1, y1, x2, y2]
        """
        orig_x1, orig_y1, orig_x2, orig_y2 = bbox
        _, orig_height,orig_width = original_size
        _, new_height, new_width = new_size
    
        x_scale = new_width / orig_width
        y_scale = new_height / orig_height
    
        new_x1 = int(orig_x1 * x_scale)
        new_y1 = int(orig_y1 * y_scale)
        new_x2 = int(orig_x2 * x_scale)
        new_y2 = int(orig_y2 * y_scale)
    
        return torch.tensor([new_x1, new_y1, new_x2, new_y2],dtype=torch.int32)

    def crop_boxes(self, boxes, top, left, height, width):
        boxes[:, [0, 2]] = boxes[:, [0, 2]] - left
        boxes[:, [1, 3]] = boxes[:, [1, 3]] - top

        boxes[:, [0, 2]] = boxes[:, [0, 2]].clamp(min=0, max=width)
        boxes[:, [1, 3]] = boxes[:, [1, 3]].clamp(min=0, max=height)
        return boxes

    def hflip_boxes(self, boxes, image_width):
        boxes[:, [0, 2]] = image_width - boxes[:, [2, 0]]
        return boxes
   
    def vflip_boxes(self, boxes, image_height):
        boxes[:, [1, 3]] = image_height - boxes[:, [3, 1]]
        return boxes
    
    def rotate_boxes(self, boxes, angle, image_height, image_width):
        # Convert the angle to radians
        angle = -angle * np.pi / 180.0
        boxes = self.rotate_polygon(boxes, angle, image_height, image_width)
        return boxes

    def rotate_polygon(self, polygon, angle, image_height, image_width):
        # Get the center of the polygon
        center = polygon.mean(axis=0)

        # Shift the polygon so that the center of the polygon is at the origin
        shifted_polygon = polygon - center

        # Rotate the polygon
        rotated_polygon = self.rotate_point(shifted_polygon, angle)

        # Shift the polygon back
        rotated_polygon += center

        return rotated_polygon
    
    def rotate_point(self, point, angle):
        # Get the rotation matrix
        rotation_matrix = np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
        return np.dot(point, rotation_matrix)


class customdl(torch.utils.data.Dataset):
    def __init__(self,data,model_type,transform=None,train_file=True,logger=None):
        self.transform=transform
        self.logger=logger
        self.folder_name=os.path.dirname(data['root'])
        self.data_type = model_type
        self.csv_data = load_any_df(data['root'],logger=self.logger)

        if self.logger:
            self.logger.info("Data file: {} loaded with mb_pandas.".format(data))
            self.logger.info("Data columns: {}".format(self.csv_data.columns))
            self.logger.info("Data will be split into train and validation according to train_file input : {}".format(train_file))
            self.logger.info("If unnamed columns are present, they will be removed.")
            self.logger.info("If duplicate rows are present, they will be removed.")
        assert 'image_path' in self.csv_data.columns, "image_path column not found in data"
        assert 'image_type' in self.csv_data.columns, "image_type column not found in data"

        #checking paths
        path_check_res= [os.path.exists(self.csv_data.image_path[i]) for i in range(len(self.csv_data))]
        self.csv_data['img_path_check'] = path_check_res ## check why test loader doesnt get this column
        self.csv_data = self.csv_data[self.csv_data['img_path_check'] == True]  # noqa: E712
        self.csv_data = self.csv_data.reset_index(drop=True)
        if logger:
            self.logger.info("Length of data after removing invalid paths: {}".format(len(self.csv_data)))

        if data['verify_image']:
            if self.logger:
                self.logger.info("Verifying images")
            verify_image_res = [verify_image(self.csv_data['image_path'].iloc[i],logger=self.logger) for i in range(len(self.csv_data))]  
            self.csv_data['img_verify'] = verify_image_res
            self.csv_data = self.csv_data[self.csv_data['img_verify'] == True]  # noqa: E712
            self.csv_data = self.csv_data.reset_index()
        else:   
            if self.logger:
                self.logger.info("Skipping image verification")

        if train_file: ## used this to differentiate between train and validation data in the data file
            try:
                if 'training' in pd.unique(self.csv_data['image_type']):
                    self.csv_data = self.csv_data[self.csv_data['image_type'] == 'training']
                if 'train' in pd.unique(self.csv_data['image_type']):   
                    self.csv_data = self.csv_data[self.csv_data['image_type'] == 'train']
            except Exception as e:
                if self.logger:
                    self.logger.info(e)
                return "image_type column (train/training) not found in data"
        else:
            try:
                if 'validation' in pd.unique(self.csv_data['image_type']):
                    self.csv_data = self.csv_data[self.csv_data['image_type'] == 'validation']
                if 'val'  in pd.unique(self.csv_data['image_type']):
                    self.csv_data = self.csv_data[self.csv_data['image_type'] == 'val']
                if 'test' in pd.unique(self.csv_data['image_type']):
                    self.csv_data = self.csv_data[self.csv_data['image_type'] == 'test']
                if 'testing' in pd.unique(self.csv_data['image_type']):
                    self.csv_data = self.csv_data[self.csv_data['image_type'] == 'testing']
            except Exception as e:
                if self.logger:
                    self.logger.info(e)
                return "image_type column (val/validation/test/testing) not found in data"

        assert 'label' in self.csv_data.columns, "label column not found in data"

        self.csv_data = check_drop_duplicates(self.csv_data,columns=['image_path'],drop=True,logger=self.logger)
        self.csv_data = remove_unnamed(self.csv_data,logger=self.logger)

        ## save label_num_map
        if os.path.exists(self.folder_name):
            output_num_labels = os.path.join(self.folder_name,'label_num_map.csv')
        else:
            output_num_labels = None
            print("Folder name not found. Please check the folder name.")
        self.csv_data = labels_num_map(self.csv_data,output_csv=output_num_labels)

        if logger:
            self.logger.info("Length of data after removing duplicates and unnamed columns: {}".format(len(self.csv_data)))
        
        if self.data_type == 'classification':
            self.label = self.csv_data['label']
            self.label_num = self.csv_data['label_num']
    
        if self.data_type == 'segmentation':
            assert 'mask_path' in self.csv_data.columns, "mask_path column not found in data"
            self.masks = self.csv_data['mask_path']
            self.label = self.csv_data['label']
            self.label_num = self.csv_data['label_num']

        if self.data_type == 'detection':
            assert 'bbox' in self.csv_data.columns, "bbox column not found in data"
            self.label = self.csv_data['label']
            self.label_num = self.csv_data['label_num']
            self.bbox = self.csv_data['bbox']

        ## save wrangled file
        if train_file:
            try:
                if os.path.exists(self.folder_name):
                    self.csv_data.to_csv(os.path.join(self.folder_name,'train_wrangled_file.csv'),index=False)
            except Exception as e:
                if self.logger:
                    self.logger.info(e)
                    self.logger.info("Could not save wrangled file. Please check the folder name.")
        else:
            try:
                if os.path.exists(self.folder_name):
                    self.csv_data.to_csv(os.path.join(self.folder_name,'val_wrangled_file.csv'),index=False)
            except Exception as e:
                if self.logger:
                    self.logger.info(e)
                    self.logger.info("Could not save wrangled file. Please check the folder name.")

    def __len__(self):
        return len(self.csv_data)
    
    def __repr__(self) -> str:
        return "self.data: {},self.transform: {},self.label: {}".format(self.csv_data,self.transform,self.label)

    def __getitem__(self,idx):
        
        img = self.csv_data['image_path'].iloc[idx]
        #img = Image.open(img)
        img = cv2.imread(img)

        if self.data_type == 'classification':
            if self.transform:
                img = self.transform(img)
            label = {}
            label['label'] = self.label_num.iloc[idx]   
            return img,label
        
        if self.data_type == 'segmentation':
            if self.transform:
                mask = cv2.imread(self.masks.iloc[idx],cv2.IMREAD_GRAYSCALE) ## considering mask is just binary class
                img,mask = self.transform(img,mask=mask)
            mask_dict={}
            mask_dict['mask'] = mask
            mask_dict['label'] = self.label_num.iloc[idx]
            
            return img,mask_dict
        
        if self.data_type == 'detection':
            bbox = eval(self.bbox.iloc[idx])
            if self.transform:
                img,bbox = self.transform(img,bbox=bbox)
            else:
                img = transforms.ToTensor()(img)
                bbox = torch.tensor([[bbox[0],bbox[1],bbox[2],bbox[3]]],dtype=torch.int32)
            bbox_dict={}
            bbox_dict['bbox'] = bbox
            bbox_dict['label'] = [self.label_num.iloc[idx]]

            return img,bbox_dict

class DataLoader(data_fetcher):
    """
    Basic dataloader for pytorch1.0
    """
    def __init__(self,yaml,logger=None) -> None:
        super().__init__(yaml, logger=logger)
        self.yaml = yaml
        self.logger = logger
        self._yaml_data = None
        self.data_dict = self.load_data_params
        self.trainloader = None
        self.testloader = None
        self.model_type = self.data_dict['model']['model_type']
        self.transformations = self.data_dict['transformation']
        self.data_params_file = self.data_dict['data']['file']

    
    def data_load(self):
        """
        return all data loaders
        """

        self.trainset = self.data_train(self.data_params_file,self.model_type, 
                                        transform=JointTransforms(self.transformations),train_file=True,logger=self.logger)
        self.testset = self.data_train(self.data_params_file,self.model_type,
                                        transform=JointTransforms(self.transformations),train_file=False,logger=self.logger)

        if self.model_type == 'classification':
            def collate_fn(batch):
                return {'image': torch.stack([b[0] for b in batch]),
                         'label': torch.tensor([b[1]['label'] for b in batch])}
        elif self.model_type == 'segmentation':
            def collate_fn(batch):
                return {'image': torch.stack([b[0] for b in batch]),
                         'mask': torch.stack([b[1]['mask'] for b in batch]), 
                        'label': torch.tensor([b[1]['label'] for b in batch])}
        elif self.model_type == 'detection':
            def collate_fn(batch):
                return {'image': list(torch.stack([b[0] for b in batch])),
                         'bbox': torch.stack([b[1]['bbox'] for b in batch]), 
                        'label': torch.tensor([b[1]['label'] for b in batch])}


        self.trainloader = torch.utils.data.DataLoader(self.trainset, 
                                                       batch_size=self.data_dict['train_params']['batch_size'], 
                                                       shuffle=self.data_dict['train_params']['shuffle'], 
                                                       num_workers=self.data_dict['train_params']['num_workers'],
                                                       worker_init_fn = lambda id: np.array(self.data_dict['train_params']['seed']),
                                                       collate_fn=collate_fn)
        self.testloader = torch.utils.data.DataLoader(self.testset, 
                                                      batch_size=self.data_dict['test_params']['batch_size'], 
                                                      shuffle=self.data_dict['test_params']['shuffle'], 
                                                      num_workers=self.data_dict['test_params']['num_workers'],
                                                      worker_init_fn = lambda id: np.array(self.data_dict['test_params']['seed']),
                                                      collate_fn=collate_fn)
        return self.trainloader,self.testloader,self.trainset,self.testset

    
    def data_train(self,data,model_type,transform=None,train_file=True,**kwargs):
        """
        get train data from yaml file
        """
        data_t = customdl(data,model_type,transform=transform,train_file=train_file,**kwargs)
        return data_t